@@ -13,8 +13,8 @@ from coolamqp.attaches.utils import Synchronized
 from coolamqp.exceptions import AMQPError, ConnectionDead
 from coolamqp.framing.definitions import ChannelOpenOk, ExchangeDeclare, \
     ExchangeDeclareOk, QueueDeclare, \
-    QueueDeclareOk, ChannelClose, QueueDelete, QueueDeleteOk
-from coolamqp.objects import Exchange, Queue, Callable
+    QueueDeclareOk, ChannelClose, QueueDelete, QueueDeleteOk, QueueBind, QueueBindOk
+from coolamqp.objects import Exchange, Queue, Callable, QueueBind as CommandQueueBind
 
 logger = logging.getLogger(__name__)
 
@@ -29,24 +29,78 @@ class Operation(object):
     This will register it's own callback. Please, call on_connection_dead when connection is broken
     to fail futures with ConnectionDead, since this object does not watch for Fails
     """
-    __slots__ = ('done', 'fut', 'declarer', 'obj', 'on_done')
+    __slots__ = ('done', 'fut', 'declarer', 'obj', 'on_done', 'parent_span', 'enqueued_span',
+                 'processing_span')
 
-    def __init__(self, declarer, obj, fut=None):
+    def __init__(self, declarer, obj, fut=None, span_parent=None, span_enqueued=None):
         self.done = False
         self.fut = fut
+        self.parent_span = span_parent
+        self.enqueued_span = span_enqueued
+        self.processing_span = None
         self.declarer = declarer
         self.obj = obj
 
         self.on_done = Callable()  # callable/0
 
+    def span_exception(self, exception):
+        if self.parent_span is not None:
+            if self.enqueued_span is None:
+                from opentracing import tags, logs
+                self.enqueued_span.set_tag(tags.ERROR, True)
+                self.enqueued_span.log_kv({logs.EVENT: tags.ERROR,
+                                             logs.ERROR_KIND: exception,
+                                             logs.ERROR_OBJECT: exception})
+                self.enqueued_span.finish()
+                self.enqueued_span = None
+
+                if self.processing_span is not None:
+                    from opentracing import tags, logs
+                    self.processing_span.set_tag(tags.ERROR, True)
+                    self.processing_span.log_kv({logs.EVENT: tags.ERROR,
+                                                 logs.ERROR_KIND: exception,
+                                                 logs.ERROR_OBJECT: exception})
+                    self.processing_span.finish()
+                    self.processing_span = None
+            if self.enqueued_span is not None:
+                self.enqueued_span.finish()
+                self.enqueued_span = None
+            self.parent_span.finish()
+
     def on_connection_dead(self):
         """To be called by declarer when our link fails"""
         if self.fut is not None:
-            self.fut.set_exception(ConnectionDead())
+            err = ConnectionDead()
+            self.span_exception(err)
+            self.fut.set_exception(err)
             self.fut = None
 
+    def span_starting(self):
+        if self.enqueued_span is not None:
+            self.enqueued_span.finish()
+            from opentracing import follows_from
+            self.processing_span = self.declarer.cluster.tracer.start_span('Declaring',
+                                                                           child_of=self.parent_span,
+                                                                           references=follows_from(self.enqueued_span))
+            self.enqueued_span = None
+
+    def span_finished(self):
+        if self.processing_span is not None:
+            self.processing_span.finish()
+            self.processing_span = None
+
+    def span_begin(self):
+        if self.enqueued_span is not None:
+            self.enqueued_span.finish()
+            from opentracing import follows_from
+            self.processing_span = self.declarer.cluster.tracer.start_span('Declaring',
+                                                                           child_of=self.parent_span,
+                                                                           references=follows_from(self.enqueued_span))
+            self.enqueued_span = None
+
     def perform(self):
         """Attempt to perform this op."""
+        self.span_begin()
         obj = self.obj
         if isinstance(obj, Exchange):
             self.declarer.method_and_watch(
@@ -61,13 +115,21 @@ class Operation(object):
                              obj.auto_delete, False, []),
                 (QueueDeclareOk, ChannelClose),
                 self._callback)
+        elif isinstance(obj, CommandQueueBind):
+            self.declarer.method_and_watch(
+                QueueBind(obj.queue, obj.exchange, obj.routing_key, False, []),
+                (QueueBindOk, ChannelClose),
+                self._callback
+            )
 
     def _callback(self, payload):
         assert not self.done
         self.done = True
         if isinstance(payload, ChannelClose):
+            err = AMQPError(payload)
+            self.span_exception(err)
             if self.fut is not None:
-                self.fut.set_exception(AMQPError(payload))
+                self.fut.set_exception(err)
                 self.fut = None
             else:
                 # something that had no Future failed. Is it in declared?
@@ -76,6 +138,11 @@ class Operation(object):
                         self.obj)  # todo access not threadsafe
                     self.declarer.on_discard(self.obj)
         else:
+            if isinstance(payload, QueueDeclareOk) and self.obj.anonymous:
+                self.obj.name = payload.queue
+                self.obj.anonymous = False
+
+            self.span_finished()
             if self.fut is not None:
                 self.fut.set_result(None)
                 self.fut = None
@@ -83,8 +150,9 @@ class Operation(object):
 
 
 class DeleteQueue(Operation):
-    def __init__(self, declarer, queue, fut):
-        super(DeleteQueue, self).__init__(declarer, queue, fut=fut)
+    def __init__(self, declarer, queue, fut, span_parent=None, span_enqueued=None):
+        super(DeleteQueue, self).__init__(declarer, queue, fut=fut, span_parent=span_parent,
+                                          span_enqueued=span_enqueued)
 
     def perform(self):
         queue = self.obj
@@ -98,8 +166,11 @@ class DeleteQueue(Operation):
         assert not self.done
         self.done = True
         if isinstance(payload, ChannelClose):
-            self.fut.set_exception(AMQPError(payload))
+            err = AMQPError(payload)
+            self.span_exception(err)
+            self.fut.set_exception(err)
         else:  # Queue.DeleteOk
+            self.span_finished()
             self.fut.set_result(None)
         self.declarer.on_operation_done()
 
@@ -111,13 +182,13 @@ class Declarer(Channeler, Synchronized):
     This also maintains a list of declared queues/exchanges, and redeclares them on each reconnect.
     """
 
-    def __init__(self):
+    def __init__(self, cluster):
         """
         Create a new declarer.
         """
         Channeler.__init__(self)
         Synchronized.__init__(self)
-
+        self.cluster = cluster
         self.declared = set()  # since Queues and Exchanges are hashable...
         # anonymous queues aren't, but we reject those
         # persistent
@@ -172,7 +243,7 @@ class Declarer(Channeler, Synchronized):
         self.in_process = None
         self._do_operations()
 
-    def delete_queue(self, queue):
+    def delete_queue(self, queue, span=None):
         """
         Delete a queue.
 
@@ -182,6 +253,7 @@ class Declarer(Channeler, Synchronized):
         If the queue is in declared consumer list, it will not be removed.
 
         :param queue: Queue instance
+        :param span: optional span, if opentracing is installed
         :return: a Future
         """
         fut = Future()
@@ -192,7 +264,7 @@ class Declarer(Channeler, Synchronized):
 
         return fut
 
-    def declare(self, obj, persistent=False):
+    def declare(self, obj, persistent=False, span=None):
         """
         Schedule to have an object declared.
 
@@ -210,12 +282,15 @@ class Declarer(Channeler, Synchronized):
 
         :param obj: Exchange or Queue instance
         :param persistent: will be redeclared upon disconnect. To remove, use "undeclare"
+        :param span: span if opentracing is installed
         :return: a Future instance
         :raise ValueError: tried to declare anonymous queue
         """
-        if isinstance(obj, Queue):
-            if obj.anonymous:
-                raise ValueError('Cannot declare anonymous queue')
+        if span is not None:
+            enqueued_span = self.cluster.tracer.start_span('Enqueued', child_of=span)
+        else:
+            span = None
+            enqueued_span = None
 
         fut = Future()
         fut.set_running_or_notify_cancel()
@@ -224,7 +299,7 @@ class Declarer(Channeler, Synchronized):
             if obj not in self.declared:
                 self.declared.add(obj)  # todo access not threadsafe
 
-        self.left_to_declare.append(Operation(self, obj, fut))
+        self.left_to_declare.append(Operation(self, obj, fut, span, enqueued_span))
         self._do_operations()
 
         return fut
