@@ -1,7 +1,4 @@
 # coding=UTF-8
-"""
-THE object you interface with
-"""
 from __future__ import print_function, absolute_import, division
 
 import logging
@@ -10,22 +7,24 @@ import typing as tp
 import warnings
 from concurrent.futures import Future
 
-import monotonic
 import six
 
 from coolamqp.attaches import Publisher, AttacheGroup, Consumer, Declarer
+from coolamqp.attaches.utils import close_future
 from coolamqp.clustering.events import ConnectionLost, MessageReceived, \
     NothingMuch, Event
 from coolamqp.clustering.single import SingleNodeReconnector
 from coolamqp.exceptions import ConnectionDead
-from coolamqp.objects import Exchange, Message, Queue, FrameLogger
+from coolamqp.objects import Exchange, Message, Queue, FrameLogger, QueueBind
 from coolamqp.uplink import ListenerThread
+from coolamqp.utils import monotonic
 
 logger = logging.getLogger(__name__)
 
-THE_POPE_OF_NOPE = NothingMuch()
+nothing_much = NothingMuch()
 
 
+# If any spans are spawn here, it's Cluster's job to finish them, except for publish()
 class Cluster(object):
     """
     Frontend for your AMQP needs.
@@ -46,6 +45,7 @@ class Cluster(object):
     :param name: name to appear in log items and prctl() for the listener thread
     :param on_blocked: callable to call when ConnectionBlocked/ConnectionUnblocked is received. It will be
         called with a value of True if connection becomes blocked, and False upon an unblock
+    :param tracer: tracer, if opentracing is installed
     """
 
     # Events you can be informed about
@@ -54,10 +54,12 @@ class Cluster(object):
 
     def __init__(self, nodes,  # type: tp.Union[NodeDefinition, tp.List[NodeDefinition]]
                  on_fail=None,  # type: tp.Optional[tp.Callable[[], None]]
-                 extra_properties=None,  # type: tp.Optional[tp.List[tp.Tuple[bytes, tp.Tuple[bytes, str]]]]
+                 extra_properties=None,
+                 # type: tp.Optional[tp.List[tp.Tuple[bytes, tp.Tuple[bytes, str]]]]
                  log_frames=None,  # type: tp.Optional[FrameLogger]
                  name=None,  # type: tp.Optional[str]
-                 on_blocked=None  # type: tp.Callable[[bool], None]
+                 on_blocked=None,  # type: tp.Callable[[bool], None],
+                 tracer=None  # type: opentracing.Traccer
                  ):
         from coolamqp.objects import NodeDefinition
         if isinstance(nodes, NodeDefinition):
@@ -66,12 +68,27 @@ class Cluster(object):
         if len(nodes) > 1:
             raise NotImplementedError(u'Multiple nodes not supported yet')
 
-        self.name = name or 'CoolAMQP'
-        self.node, = nodes
+        if tracer is not None:
+            try:
+                import opentracing
+            except ImportError:
+                raise RuntimeError('tracer given, but opentracing is not installed!')
+
+        self.started = False            # type: bool
+        self.tracer = tracer
+        self.name = name or 'CoolAMQP'  # type: str
+        self.node, = nodes              # type: NodeDefinition
         self.extra_properties = extra_properties
-        self.log_frames = log_frames
-        self.on_blocked = on_blocked
-        self.connected = False
+        self.log_frames = log_frames    # type: tp.Optional[FrameLogger]
+        self.on_blocked = on_blocked    # type: tp.Optional[tp.Callable[[bool], None]]
+        self.connected = False          # type: bool
+        self.listener = None            # type: BaseListener
+        self.attache_group = None       # type: AttacheGroup
+        self.events = None              # type: six.moves.queue.Queue
+        self.snr = None                 # type: SingleNodeReconnector
+        self.pub_tr = None              # type: Publisher
+        self.pub_na = None              # type: Publisher
+        self.decl = None                # type: Declarer
 
         if on_fail is not None:
             def decorated():
@@ -82,35 +99,73 @@ class Cluster(object):
         else:
             self.on_fail = None
 
+    def bind(self, queue, exchange, routing_key, persistent=False, span=None):
+        """
+        Bind a queue to an exchange
+        """
+        if span is not None:
+            child_span = self._make_span('bind', span)
+        else:
+            child_span = None
+        fut = self.decl.declare(QueueBind(queue, exchange, routing_key),
+                                persistent=persistent,
+                                span=child_span)
+        return close_future(fut, child_span)
+
     def declare(self, obj,  # type: tp.Union[Queue, Exchange]
-                persistent=False  # type: bool
+                persistent=False,  # type: bool
+                span=None  # type: tp.Optional[opentracing.Span]
                 ):  # type: (...) -> concurrent.futures.Future
         """
         Declare a Queue/Exchange
 
         :param obj: Queue/Exchange object
         :param persistent: should it be redefined upon reconnect?
+        :param span: optional parent span, if opentracing is installed
         :return: Future
         """
-        return self.decl.declare(obj, persistent=persistent)
+        if span is not None:
+            child_span = self._make_span('declare', span)
+        else:
+            child_span = None
+        fut = self.decl.declare(obj, persistent=persistent, span=child_span)
+        return close_future(fut, child_span)
 
-    def drain(self, timeout):  # type: (float) -> Event
+    def drain(self, timeout, span=None):  # type: (float) -> Event
         """
         Return an Event.
 
         :param timeout: time to wait for an event. 0 means return immediately. None means block forever
+        :param span: optional parent span, if opentracing is installed
         :return: an Event instance. NothingMuch is returned when there's nothing within a given timoeout
         """
-        try:
-            if timeout == 0:
-                return self.events.get_nowait()
-            else:
-                return self.events.get(True, timeout)
-        except six.moves.queue.Empty:
-            return THE_POPE_OF_NOPE
 
-    def consume(self, queue, on_message=None, *args, **kwargs):
-        # type: (Queue, tp.Callable[[MessageReceived], None] -> tp.Tuple[Consumer, Future]
+        def fetch():
+            try:
+                if timeout == 0:
+                    return self.events.get_nowait()
+                else:
+                    return self.events.get(True, timeout)
+            except six.moves.queue.Empty:
+                return nothing_much
+
+        if span is not None:
+            from opentracing import tags
+            parent_span = self.tracer.start_active_span('AMQP call',
+                                                        child_of=span,
+                                                        tags={
+                                                            tags.SPAN_KIND: tags.SPAN_KIND_RPC_CLIENT,
+                                                            tags.DATABASE_TYPE: 'amqp',
+                                                            tags.DATABASE_STATEMENT: 'drain'
+                                                        })
+
+            with parent_span:
+                return fetch()
+        else:
+            return fetch()
+
+    def consume(self, queue, on_message=None, span=None, *args, **kwargs):
+        # type: (Queue, tp.Callable[[MessageReceived], None]) -> tp.Tuple[Consumer, Future]
         """
         Start consuming from a queue.
 
@@ -123,16 +178,21 @@ class Cluster(object):
             Note that name of anonymous queue might change at any time!
         :param on_message: callable that will process incoming messages
                            if you leave it at None, messages will be .put into self.events
+        :param span: optional span, if opentracing is installed
         :return: a tuple (Consumer instance, and a Future), that tells, when consumer is ready
         """
+        if span is not None:
+            child_span = self._make_span('consume', span)
+        else:
+            child_span = None
         fut = Future()
         fut.set_running_or_notify_cancel()  # it's running right now
         on_message = on_message or (
-            lambda rmsg: self.events.put_nowait(MessageReceived(rmsg)))
-        con = Consumer(queue, on_message, future_to_notify=fut, *args,
+            lambda msg: self.events.put_nowait(MessageReceived(msg)))
+        con = Consumer(queue, on_message, future_to_notify=fut, span=span, *args,
                        **kwargs)
         self.attache_group.add(con)
-        return con, fut
+        return con, close_future(fut, child_span)
 
     def delete_queue(self, queue):  # type: (coolamqp.objects.Queue) -> Future
         """
@@ -143,11 +203,26 @@ class Cluster(object):
         """
         return self.decl.delete_queue(queue)
 
+    def _make_span(self, call, span):
+        try:
+            from opentracing import tags
+        except ImportError:
+            pass
+        else:
+            return self.tracer.start_span('AMQP call',
+                                          child_of=span,
+                                          tags={
+                                              tags.SPAN_KIND: tags.SPAN_KIND_RPC_CLIENT,
+                                              tags.DATABASE_TYPE: 'amqp',
+                                              tags.DATABASE_STATEMENT: call
+                                          })
+
     def publish(self, message,  # type: Message
                 exchange=None,  # type: tp.Union[Exchange, str, bytes]
                 routing_key=u'',  # type: tp.Union[str, bytes]
                 tx=None,  # type: tp.Optional[bool]
-                confirm=None  # type: tp.Optional[bool]
+                confirm=None,  # type: tp.Optional[bool]
+                span=None  # type: tp.Optional[opentracing.Span]
                 ):  # type: (...) -> tp.Optional[Future]
         """
         Publish a message.
@@ -161,8 +236,12 @@ class Cluster(object):
                         Note that if tx if False, and message cannot be delivered to broker at once,
                         it will be discarded
         :param tx: deprecated, alias for confirm
+        :param span: optionally, current span, if opentracing is installed
         :return: Future to be finished on completion or None, is confirm/tx was not chosen
         """
+        if self.tracer is not None:
+            span = self._make_span('publish', span)
+
         if isinstance(exchange, Exchange):
             exchange = exchange.name.encode('utf8')
         elif exchange is None:
@@ -185,14 +264,16 @@ class Cluster(object):
             tx = False
 
         try:
-            return (self.pub_tr if tx else self.pub_na).publish(message,
-                                                                exchange,
-                                                                routing_key)
+            if tx:
+                clb = self.pub_tr
+            else:
+                clb = self.pub_na
+            return clb.publish(message, exchange, routing_key, span)
         except Publisher.UnusablePublisher:
             raise NotImplementedError(
                 u'Sorry, this functionality is not yet implemented!')
 
-    def start(self, wait=True, timeout=10.0, log_frames=None):  # type: (bool, float, bool) -> None
+    def start(self, wait=True, timeout=10.0):  # type: (bool, float, bool) -> None
         """
         Connect to broker. Initialize Cluster.
 
@@ -204,15 +285,10 @@ class Cluster(object):
                         ConnectionDead error will be raised
         :raise RuntimeError: called more than once
         :raise ConnectionDead: failed to connect within timeout
-        :param log_frames: whether to keep a log of sent/received frames in self.log_frames
         """
-
-        try:
-            self.listener
-        except AttributeError:
-            pass
-        else:
+        if self.started:
             raise RuntimeError(u'[%s] This was already called!' % (self.name,))
+        self.started = True
 
         self.listener = ListenerThread(name=self.name)
 
@@ -222,7 +298,7 @@ class Cluster(object):
 
         self.snr = SingleNodeReconnector(self.node, self.attache_group,
                                          self.listener, self.extra_properties,
-                                         log_frames, self.name)
+                                         self.log_frames, self.name)
         self.snr.on_fail.add(lambda: self.events.put_nowait(ConnectionLost()))
         if self.on_fail is not None:
             self.snr.on_fail.add(self.on_fail)
@@ -231,9 +307,9 @@ class Cluster(object):
             self.snr.on_blocked.add(self.on_blocked)
 
         # Spawn a transactional publisher and a noack publisher
-        self.pub_tr = Publisher(Publisher.MODE_CNPUB, cluster_to_set_connected_upon_first_connect=self)
-        self.pub_na = Publisher(Publisher.MODE_NOACK)
-        self.decl = Declarer()
+        self.pub_tr = Publisher(Publisher.MODE_CNPUB, self)
+        self.pub_na = Publisher(Publisher.MODE_NOACK, self)
+        self.decl = Declarer(self)
 
         self.attache_group.add(self.pub_tr)
         self.attache_group.add(self.pub_na)
@@ -245,8 +321,8 @@ class Cluster(object):
 
         if wait:
             # this is only going to take a short amount of time, so we're fine with polling
-            start_at = monotonic.monotonic()
-            while not self.connected and monotonic.monotonic() - start_at < timeout:
+            start_at = monotonic()
+            while not self.connected and monotonic() - start_at < timeout:
                 time.sleep(0.1)
             if not self.connected:
                 raise ConnectionDead(
@@ -260,9 +336,7 @@ class Cluster(object):
         :raise RuntimeError: if called without start() being called first
         """
         self.connected = False
-        try:
-            self.listener
-        except AttributeError:
+        if not self.started:
             raise RuntimeError(u'shutdown without start')
 
         logger.info('[%s] Commencing shutdown', self.name)
